from datetime import datetime
from time import time

from dearpygui import dearpygui as dpg
from keyboard import add_hotkey

from leads import *
from leads.comm import *
from leads_dashboard import *
from leads_vec.__version__ import __version__
from leads_vec.config import Config


class CustomRuntimeData(RuntimeData):
    m1_mode: int = 0
    m3_mode: int = 0


def main(main_controller: Controller, config: Config) -> int:
    context = Leads(srw_mode=config.srw_mode)
    rd = CustomRuntimeData()

    class CustomCallback(Callback):
        def on_fail(self, service: Service, error: Exception) -> None:
            rd.comm = None

        def on_receive(self, service: Service, msg: bytes) -> None:
            print(msg)

    rd.comm = start_client(config.comm_addr, create_client(config.comm_port, CustomCallback()), True)

    def switch_m1_mode():
        rd.m1_mode = (rd.m1_mode + 1) % 2

    def switch_m3_mode():
        rd.m3_mode = (rd.m3_mode + 1) % 3

    def switch_dtcs():
        context.set_dtcs(not (dtcs_enabled := context.is_dtcs_enabled()))
        dpg.set_item_label("dtcs", f"DTCS {'OFF' if dtcs_enabled else 'ON'}")

    add_hotkey("1", switch_dtcs)

    def switch_abs():
        context.set_abs(not (abs_enabled := context.is_abs_enabled()))
        dpg.set_item_label("abs", f"ABS {'OFF' if abs_enabled else 'ON'}")

    add_hotkey("2", switch_abs)

    def switch_ebi():
        context.set_ebi(not (ebi_enabled := context.is_ebi_enabled()))
        dpg.set_item_label("ebi", f"EBI {'OFF' if ebi_enabled else 'ON'}")

    add_hotkey("3", switch_ebi)

    def switch_atbs():
        context.set_atbs(not (atbs_enabled := context.is_atbs_enabled()))
        dpg.set_item_label("atbs", f"ATBS {'OFF' if atbs_enabled else 'ON'}")

    add_hotkey("4", switch_atbs)

    def render():
        with dpg.table(header_row=False):
            dpg.add_table_column()
            dpg.add_table_column()
            dpg.add_table_column()
            with dpg.table_row():
                dpg.bind_item_font(dpg.add_button(label="", tag="m1", width=-1, height=140), BODY2)
                dpg.set_item_callback("m1", switch_m1_mode)
                dpg.bind_item_font(dpg.add_button(label="0", tag="m2", width=-1, height=140), H1)
                dpg.bind_item_font(dpg.add_button(label="0.0v", tag="m3", width=-1, height=140), H1)
                dpg.set_item_callback("m3", switch_m3_mode)
        with dpg.table(header_row=False):
            dpg.add_table_column()
            dpg.add_table_column()
            dpg.add_table_column()
            dpg.add_table_column()
            dpg.add_table_column()
            with dpg.table_row():
                dpg.bind_item_font(dpg.add_text("DTCS READY", color=(0, 255, 0), tag="dtcs_status"), BODY)
                dpg.bind_item_font(dpg.add_text("ABS READY", color=(0, 255, 0), tag="abs_status"), BODY)
                dpg.bind_item_font(dpg.add_text("EBI READY", color=(0, 255, 0), tag="ebi_status"), BODY)
                dpg.bind_item_font(dpg.add_text("ATBS READY", color=(0, 255, 0), tag="atbs_status"), BODY)
                dpg.bind_item_font(dpg.add_text("COMM ONLINE", tag="comm_status"), BODY)
            with dpg.table_row():
                dpg.bind_item_font(dpg.add_button(label="DTCS ON",
                                                  tag="dtcs",
                                                  width=-1,
                                                  callback=switch_dtcs), BODY)
                dpg.bind_item_font(dpg.add_button(label="ABS ON",
                                                  tag="abs",
                                                  width=-1,
                                                  callback=switch_abs), BODY)
                dpg.bind_item_font(dpg.add_button(label="EBI ON",
                                                  tag="ebi",
                                                  width=-1,
                                                  callback=switch_ebi), BODY)
                dpg.bind_item_font(dpg.add_button(label="ATBS ON",
                                                  tag="atbs",
                                                  width=-1,
                                                  callback=switch_atbs), BODY)

    class CustomListener(EventListener):
        def on_push(self, e: DataPushedEvent) -> None:
            try:
                rd.comm_notify(e.data)
            except IOError:
                rd.comm_kill()
                rd.comm = None
                dpg.set_value("comm_status", "COMM OFFLINE")

        def on_update(self, e: UpdateEvent) -> None:
            duration = int(time()) - rd.start_time
            if rd.m1_mode == 0:
                dpg.set_item_label("m1", "LAP TIME\n\nLAP1 9s\nLAP2 11s\nLAP3 10s")
            elif rd.m1_mode == 1:
                dpg.set_item_label("m1",
                                   f"VeC {__version__.upper()}\n\n"
                                   f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}\n"
                                   f"{duration // 60} MIN {duration % 60} SEC\n\n"
                                   f"{'SRW MODE' if config.srw_mode else 'DRW MODE'}\n"
                                   f"ANALYSIS RATE: {int(1 / config.analysis_rate)} TPS\n"
                                   f"UPDATE RATE: {int(1 / config.update_rate)} TPS")
            dpg.set_item_label("m2", f"{int(context.data().front_wheel_speed)}")
            if rd.m3_mode == 0:
                dpg.bind_item_font("m3", H1)
                dpg.set_item_label("m3", "0.0v")
            elif rd.m3_mode == 1:
                dpg.bind_item_font("m3", BODY)
                dpg.set_item_label("m3", "G Force")
            else:
                dpg.bind_item_font("m3", BODY)
                dpg.set_item_label("m3", "SPEED TREND")
            dpg.set_value("comm_status", "COMM ONLINE" if rd.comm else "COMM OFFLINE")

        def on_intervene(self, e: InterventionEvent) -> None:
            dpg.set_value(e.system.lower() + "_status", e.system + " INTERVENED")

        def post_intervene(self, e: InterventionEvent) -> None:
            dpg.set_value(e.system.lower() + "_status", e.system + " READY")

        def on_suspend(self, e: SuspensionEvent) -> None:
            dpg.set_value(e.system.lower() + "_status", e.system + " SUSPENDED")

        def post_suspend(self, e: SuspensionEvent) -> None:
            dpg.set_value(e.system.lower() + "_status", e.system + " READY")

    context.set_event_listener(CustomListener())
    start(render, context, main_controller, config.analysis_rate, config.update_rate, rd)
    rd.comm_kill()
    return 0



from sys import argv as _argv, exit as _exit

from leads_vec.config import load_config as _load_config, DEFAULT_CONFIG as _DEFAULT_CONFIG


def get_arg(args: list[str], name: str) -> str | None:
    for arg in args:
        if arg.startswith(name + "="):
            return arg[len(name) + 1:]


if __name__ == '__main__':
    try:
        from leads_emulation import SRWRandom as _Controller
    except ImportError:
        raise ImportError("At least one adapter has to be installed")
    cfg_arg = get_arg(_argv, "config")
    config = _load_config(cfg_arg) if cfg_arg else _DEFAULT_CONFIG

    if "--remote" in _argv:
        from leads_vec.remote import remote as _main

        _exit(_main(config))
    else:
        from leads_vec.cli import main as _main

        _exit(_main(_Controller("main"), config))



from importlib.util import find_spec as _find_spec

if not _find_spec("dearpygui"):
    raise ImportError("Please install `dearpygui` to run this module\n>>> pip install dearpygui")
if not _find_spec("keyboard"):
    raise ImportError("Please install `keyboard` to run this module\n>>> pip install keyboard")



from time import sleep as _sleep
from typing import Callable as _Callable, TypeVar as _TypeVar

from dearpygui import dearpygui as _dpg

from leads import Leads as _Leads, Controller as _Controller
from leads.comm import Server as _Server, create_server as _create_server
from leads_dashboard.fonts import load_font
from leads_dashboard.runtime import RuntimeData

_dpg.create_context()

H1 = load_font(size=100)
H2 = load_font(size=60)
H3 = load_font(size=40)
H4 = load_font(size=30)
H5 = load_font(size=20)
BODY = load_font(size=10)
BODY2 = load_font(size=6)

_dpg.create_viewport(title="LEADS")
_dpg.setup_dearpygui()


T = _TypeVar("T")


def start(render: _Callable[[], None],
          context: _Leads[T],
          main_controller: _Controller[T],
          analysis_rate: float,
          update_rate: float,
          runtime_data: RuntimeData) -> None:
    with _dpg.window(tag="main",
                     label="LEADS",
                     no_title_bar=True,
                     no_scrollbar=True,
                     menubar=False,
                     no_close=True,
                     no_background=True):
        render()
    _dpg.show_viewport()
    _dpg.set_primary_window("main", True)
    runtime_data.frame_counter = 0
    while _dpg.is_dearpygui_running():
        _sleep(analysis_rate)
        context.push(main_controller.collect_all())
        if runtime_data.frame_counter % (update_rate / analysis_rate) == 0:
            context.update()
        runtime_data.frame_counter += 1
        _dpg.render_dearpygui_frame()
    _dpg.destroy_context()


def start_comm_server(render: _Callable[[], None], server: _Server = _create_server()) -> None:
    with _dpg.window(tag="main",
                     label="LEADS Comm",
                     no_title_bar=True,
                     no_scrollbar=True,
                     menubar=False,
                     no_close=True,
                     no_background=True):
        render()
    _dpg.show_viewport()
    _dpg.set_primary_window("main", True)
    server.start(True)
    _dpg.start_dearpygui()
    _dpg.destroy_context()
    server.kill()



from collections import deque
from json import loads
from os import mkdir
from os.path import exists

from dearpygui import dearpygui as dpg
from numpy import trapz, concatenate, array

from leads.comm import *
from leads.utils import *
from leads_dashboard import *
from leads_vec.config import Config


def render():
    dpg.bind_item_font(dpg.add_text("", tag="speed"), H2)
    dpg.bind_item_font(dpg.add_text("", tag="displacement"), H2)
    dpg.add_simple_plot(label="Speed Trend", tag="speed_seq", height=300)


def integrate_speed2displacement(x: DataPersistence, y: DataPersistence) -> float:
    return 0 if len(x) < 2 else trapz(y.to_list() + y.get_chunk(),
                                      concatenate((array(a := x.to_list()) - a[0], array(x.get_chunk()) - a[0])) / 3600,
                                      dx=.001)


def remote(config: Config) -> int:
    if not exists(config.data_dir):
        mkdir(config.data_dir)
        print(f"Data dir \"{config.data_dir}\" created")

    class CustomCallback(Callback):
        speed_seq: deque = deque(maxlen=512)
        speed_record: DataPersistence = DataPersistence(config.data_dir + "/speed.csv", max_size=256)
        time_stamp_record: DataPersistence = DataPersistence(config.data_dir + "/time_stamp.csv", max_size=256)

        def on_initialize(self, service: Service) -> None:
            print(f"Comm server started on port {service.port()}")

        def on_fail(self, service: Service, error: Exception) -> None:
            print(error)

        def on_receive(self, service: Service, msg: bytes) -> None:
            data = loads(msg.decode())
            self.time_stamp_record.append(data["t"])
            front_wheel_speed = data["front_wheel_speed"]
            self.speed_seq.append(front_wheel_speed)
            self.speed_record.append(front_wheel_speed)
            dpg.set_value("speed", f"FWS: {int(front_wheel_speed)} KM / H")
            dpg.set_value("displacement",
                          f"{int(integrate_speed2displacement(self.time_stamp_record, self.speed_record))} M")
            dpg.set_value("speed_seq", list(self.speed_seq))

        def on_disconnect(self, service: Service) -> None:
            self.speed_record.close()

    start_comm_server(render, create_server(config.comm_port, CustomCallback()))
    return 0
